"""
***************************************************************************
    MessageBarProgress.py
    ---------------------
    Date                 : April 2013
    Copyright            : (C) 2013 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "April 2013"
__copyright__ = "(C) 2013, Victor Olaya"

from qgis.core import Qgis, QgsProcessingFeedback
from qgis.PyQt.QtCore import QCoreApplication, Qt
from qgis.PyQt.QtWidgets import QProgressBar
from qgis.utils import iface

from processing.gui.MessageDialog import MessageDialog


class MessageBarProgress(QgsProcessingFeedback):
    def __init__(self, algname=None):
        QgsProcessingFeedback.__init__(self)

        self.msg = []
        self.progressMessageBar = iface.messageBar().createMessage(
            self.tr("Executing algorithm <i>{}</i>").format(algname if algname else "")
        )
        self.progress = QProgressBar()
        self.progressChanged.connect(self.set_progress_bar_value)
        self.progress.setMaximum(100)
        self.progress.setAlignment(
            Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
        )
        self.progressMessageBar.layout().addWidget(self.progress)
        self.message_bar_item = iface.messageBar().pushWidget(
            self.progressMessageBar, Qgis.MessageLevel.Info
        )

    def set_progress_bar_value(self, progress: float):
        """
        Sets the progress bar value to a rounded int of the algorithm's
        progress
        """
        self.progress.setValue(int(progress))

    def reportError(self, msg, fatalError=False):
        self.msg.append(msg)

    def close(self):
        if self.msg:
            dlg = MessageDialog()
            dlg.setTitle(
                QCoreApplication.translate(
                    "MessageBarProgress", "Problem executing algorithm"
                )
            )
            dlg.setMessage("<br>".join(self.msg))
            dlg.exec()
        iface.messageBar().popWidget(self.message_bar_item)

    def tr(self, string, context=""):
        if context == "":
            context = "MessageBarProgress"
        return QCoreApplication.translate(context, string)



"""
***************************************************************************
    BatchOutputSelectionPanel.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os
import re

from qgis.core import (
    QgsApplication,
    QgsMapLayer,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterFolderDestination,
    QgsProcessingParameterMapLayer,
    QgsProcessingParameterMatrix,
    QgsProcessingParameterMultipleLayers,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterVectorLayer,
    QgsProject,
    QgsSettings,
)
from qgis.PyQt.QtGui import QAction
from qgis.PyQt.QtWidgets import (
    QFileDialog,
    QHBoxLayout,
    QLineEdit,
    QMenu,
    QPushButton,
    QSizePolicy,
    QWidget,
)

from processing.gui.AutofillDialog import AutofillDialog


class BatchOutputSelectionPanel(QWidget):
    def __init__(self, output, alg, row, col, panel):
        super().__init__(None)
        self.alg = alg
        self.row = row
        self.col = col
        self.output = output
        self.panel = panel
        self.table = self.panel.tblParameters
        self.horizontalLayout = QHBoxLayout(self)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setMargin(0)
        self.text = QLineEdit()
        self.text.setText("")
        self.text.setMinimumWidth(300)
        self.text.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        self.horizontalLayout.addWidget(self.text)
        self.pushButton = QPushButton()
        self.pushButton.setText("…")

        self.actionTemporaryOutput = QAction(
            self.tr("Create Temporary Layer"),
            self,
        )
        self.actionTemporaryOutput.setCheckable(True)
        self.actionTemporaryOutput.changed.connect(self.onTemporaryOutputChanged)

        if isinstance(self.output, QgsProcessingParameterFolderDestination):
            self.pushButton.clicked.connect(self.showSelectionDialog)
        else:
            self.menu = QMenu(self.pushButton)
            self.menu.addAction(
                self.tr("Select File/Folder…"), self.showSelectionDialog
            )
            self.menu.addAction(self.actionTemporaryOutput)
            self.pushButton.setMenu(self.menu)

        self.horizontalLayout.addWidget(self.pushButton)
        self.setLayout(self.horizontalLayout)

        self.actionTemporaryOutputIcon = QAction(
            QgsApplication.getThemeIcon("/mActionCreateMemory.svg"),
            self.tr("Temporary Output"),
            self,
        )

    def showSelectionDialog(self):
        if isinstance(self.output, QgsProcessingParameterFolderDestination):
            self.selectDirectory()
            return

        filefilter = self.output.createFileFilter()
        settings = QgsSettings()
        if settings.contains("/Processing/LastBatchOutputPath"):
            path = str(settings.value("/Processing/LastBatchOutputPath"))
        else:
            path = ""
        filename, selectedFileFilter = QFileDialog.getSaveFileName(
            self, self.tr("Save File"), path, filefilter
        )
        if filename:
            self.actionTemporaryOutput.setChecked(False)

            if not filename.lower().endswith(
                tuple(re.findall("\\*(\\.[a-z]{1,10})", filefilter))
            ):
                ext = re.search("\\*(\\.[a-z]{1,10})", selectedFileFilter)
                if ext:
                    filename += ext.group(1)
            settings.setValue(
                "/Processing/LastBatchOutputPath", os.path.dirname(filename)
            )
            dlg = AutofillDialog(self.alg)
            dlg.exec()
            if dlg.mode is not None:
                if dlg.mode == AutofillDialog.DO_NOT_AUTOFILL:
                    self.table.cellWidget(self.row, self.col).setValue(filename)
                elif dlg.mode == AutofillDialog.FILL_WITH_NUMBERS:
                    n = self.table.rowCount() - self.row
                    for i in range(n):
                        name = (
                            filename[: filename.rfind(".")]
                            + str(i + 1)
                            + filename[filename.rfind(".") :]
                        )
                        self.table.cellWidget(i + self.row, self.col).setValue(name)
                        self.table.cellWidget(
                            i + self.row, self.col
                        ).setTemporaryOutput(False)
                elif dlg.mode == AutofillDialog.FILL_WITH_PARAMETER:
                    for row in range(self.row, self.table.rowCount()):
                        v = self.panel.valueForParameter(row - 1, dlg.param_name)
                        param = self.alg.parameterDefinition(dlg.param_name)
                        if isinstance(
                            param,
                            (
                                QgsProcessingParameterRasterLayer,
                                QgsProcessingParameterFeatureSource,
                                QgsProcessingParameterVectorLayer,
                                QgsProcessingParameterMultipleLayers,
                                QgsProcessingParameterMapLayer,
                            ),
                        ):
                            if isinstance(v, QgsMapLayer):
                                s = v.name()
                            else:
                                if v in QgsProject.instance().mapLayers():
                                    layer = QgsProject.instance().mapLayer(v)
                                    # value is a layer ID, but we'd prefer to show a layer name if it's unique in the project
                                    if (
                                        len(
                                            [
                                                l
                                                for _, l in QgsProject.instance()
                                                .mapLayers()
                                                .items()
                                                if l.name().lower()
                                                == layer.name().lower()
                                            ]
                                        )
                                        == 1
                                    ):
                                        s = layer.name()
                                    else:
                                        # otherwise fall back to layer id
                                        s = v
                                else:
                                    # else try to use file base name
                                    # TODO: this is bad for database sources!!
                                    s = os.path.basename(v)
                                    s = os.path.splitext(s)[0]
                        elif isinstance(param, QgsProcessingParameterBoolean):
                            s = "true" if v else "false"
                        elif isinstance(param, QgsProcessingParameterEnum):
                            s = param.options()[v]
                        else:
                            s = str(v)
                        name = (
                            filename[: filename.rfind(".")]
                            + s
                            + filename[filename.rfind(".") :]
                        )
                        self.table.cellWidget(row, self.col).setValue(name)
                        self.table.cellWidget(row, self.col).setTemporaryOutput(False)

    def selectDirectory(self):

        settings = QgsSettings()
        if settings.contains("/Processing/LastBatchOutputPath"):
            lastDir = str(settings.value("/Processing/LastBatchOutputPath"))
        else:
            lastDir = ""

        dirName = QFileDialog.getExistingDirectory(
            self, self.tr("Output Directory"), lastDir, QFileDialog.Option.ShowDirsOnly
        )

        if dirName:
            self.table.cellWidget(self.row, self.col).setValue(dirName)
            settings.setValue("/Processing/LastBatchOutputPath", dirName)

    def setValue(self, text):
        return self.text.setText(text)

    def getValue(self):
        return str(self.text.text())

    def outputName(self) -> str:
        return self.text.text()

    def isTemporaryOutput(self) -> bool:
        return self.actionTemporaryOutput.isChecked()

    def onTemporaryOutputChanged(self) -> None:
        if self.actionTemporaryOutput.isChecked():
            self.text.addAction(
                self.actionTemporaryOutputIcon, QLineEdit.ActionPosition.LeadingPosition
            )
        else:
            self.text.removeAction(self.actionTemporaryOutputIcon)
        self.text.repaint()

    def setTemporaryOutput(self, temporary: bool) -> None:
        self.actionTemporaryOutput.setChecked(temporary)



"""
***************************************************************************
    InterpolationDataWidget.py
    ---------------------
    Date                 : December 2016
    Copyright            : (C) 2016 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "December 2016"
__copyright__ = "(C) 2016, Alexander Bruy"

import os
from typing import Optional, Union

from qgis.analysis import QgsInterpolator
from qgis.core import (
    Qgis,
    QgsApplication,
    QgsCoordinateReferenceSystem,
    QgsFieldProxyModel,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterNumber,
    QgsProcessingUtils,
    QgsRectangle,
    QgsReferencedRectangle,
    QgsVectorLayer,
    QgsWkbTypes,
)
from qgis.gui import QgsDoubleSpinBox
from qgis.PyQt import uic
from qgis.PyQt.QtCore import pyqtSignal
from qgis.PyQt.QtWidgets import QComboBox, QTreeWidgetItem

from processing.gui.wrappers import DIALOG_STANDARD, WidgetWrapper
from processing.tools import dataobjects

pluginPath = os.path.dirname(__file__)


class ParameterInterpolationData(QgsProcessingParameterDefinition):
    def __init__(self, name="", description=""):
        super().__init__(name, description)
        self.setMetadata(
            {
                "widget_wrapper": "processing.algs.qgis.ui.InterpolationWidgets.InterpolationDataWidgetWrapper"
            }
        )

    def type(self):
        return "idw_interpolation_data"

    def clone(self):
        return ParameterInterpolationData(self.name(), self.description())

    @staticmethod
    def parseValue(value):
        if value is None:
            return None

        if value == "":
            return None

        if isinstance(value, str):
            return value if value != "" else None
        else:
            return ParameterInterpolationData.dataToString(value)

    @staticmethod
    def dataToString(data):
        s = ""
        for c in data:
            s += f"{c[0]}::~::{c[1]}::~::{c[2]:d}::~::{c[3]:d};"
        return s[:-1]


WIDGET, BASE = uic.loadUiType(
    os.path.join(pluginPath, "interpolationdatawidgetbase.ui")
)


class InterpolationDataWidget(BASE, WIDGET):
    hasChanged = pyqtSignal()

    def __init__(self):
        super().__init__(None)
        self.setupUi(self)

        self.btnAdd.setIcon(QgsApplication.getThemeIcon("/symbologyAdd.svg"))
        self.btnRemove.setIcon(QgsApplication.getThemeIcon("/symbologyRemove.svg"))

        self.btnAdd.clicked.connect(self.addLayer)
        self.btnRemove.clicked.connect(self.removeLayer)

        self.cmbLayers.layerChanged.connect(self.layerChanged)
        self.cmbLayers.setFilters(Qgis.LayerFilter.VectorLayer)
        self.cmbFields.setFilters(QgsFieldProxyModel.Filter.Numeric)

        self.layerChanged(self.cmbLayers.currentLayer())

    def addLayer(self):
        layer = self.cmbLayers.currentLayer()
        if not layer:
            return

        attribute = ""
        if self.chkUseZCoordinate.isChecked():
            attribute = "Z_COORD"
        else:
            attribute = self.cmbFields.currentField()

        self._addLayerData(layer.name(), attribute)
        self.hasChanged.emit()

    def removeLayer(self):
        item = self.layersTree.currentItem()
        if not item:
            return
        self.layersTree.invisibleRootItem().removeChild(item)
        self.hasChanged.emit()

    def layerChanged(self, layer: Optional[QgsVectorLayer]):
        self.chkUseZCoordinate.setEnabled(False)
        self.chkUseZCoordinate.setChecked(False)

        if layer is None or not layer.isValid():
            return

        provider = layer.dataProvider()
        if not provider:
            return

        if QgsWkbTypes.hasZ(provider.wkbType()):
            self.chkUseZCoordinate.setEnabled(True)

        self.cmbFields.setLayer(layer)

    def _addLayerData(self, layerName: str, attribute: str):
        item = QTreeWidgetItem()
        item.setText(0, layerName)
        item.setText(1, attribute)
        self.layersTree.addTopLevelItem(item)

        comboBox = QComboBox()
        comboBox.addItem(self.tr("Points"))
        comboBox.addItem(self.tr("Structure lines"))
        comboBox.addItem(self.tr("Break lines"))
        comboBox.setCurrentIndex(0)
        self.layersTree.setItemWidget(item, 2, comboBox)

    def setValue(self, value: str):
        self.layersTree.clear()
        rows = value.split("::|::")
        for i, r in enumerate(rows):
            v = r.split("::~::")
            layer = QgsProcessingUtils.mapLayerFromString(
                v[0], dataobjects.createContext()
            )
            if layer is None or not layer.isValid():
                continue

            field_index = int(v[2])
            if field_index == -1:
                field_name = "Z_COORD"
            else:
                field_name = layer.fields().at(field_index).name()

            self._addLayerData(v[0], field_name)

            comboBox = self.layersTree.itemWidget(self.layersTree.topLevelItem(i), 2)
            comboBox.setCurrentIndex(int(v[3]))

        self.hasChanged.emit()

    def value(self) -> str:
        layers = ""
        context = dataobjects.createContext()
        for i in range(self.layersTree.topLevelItemCount()):
            item = self.layersTree.topLevelItem(i)
            if item:
                layerName = item.text(0)
                layer = QgsProcessingUtils.mapLayerFromString(layerName, context)
                if not layer:
                    continue

                provider = layer.dataProvider()
                if not provider:
                    continue

                interpolationAttribute = item.text(1)
                interpolationSource = QgsInterpolator.ValueSource.Attribute
                if interpolationAttribute == "Z_COORD":
                    interpolationSource = QgsInterpolator.ValueSource.Z
                    fieldIndex = -1
                else:
                    fieldIndex = layer.fields().indexFromName(interpolationAttribute)

                comboBox = self.layersTree.itemWidget(
                    self.layersTree.topLevelItem(i), 2
                )
                inputTypeName = comboBox.currentText()
                if inputTypeName == self.tr("Points"):
                    inputType = QgsInterpolator.SourceType.Points
                elif inputTypeName == self.tr("Structure lines"):
                    inputType = QgsInterpolator.SourceType.StructureLines
                else:
                    inputType = QgsInterpolator.SourceType.BreakLines

                layers += f"{layer.source()}::~::{interpolationSource:d}::~::{fieldIndex:d}::~::{inputType:d}::|::"
        return layers[: -len("::|::")]


class InterpolationDataWidgetWrapper(WidgetWrapper):
    def createWidget(self):
        widget = InterpolationDataWidget()
        widget.hasChanged.connect(lambda: self.widgetValueHasChanged.emit(self))
        return widget

    def setValue(self, value):
        self.widget.setValue(value)

    def value(self):
        return self.widget.value()


class ParameterPixelSize(QgsProcessingParameterNumber):
    def __init__(
        self,
        name="",
        description="",
        layersData=None,
        extent=None,
        minValue=None,
        default=None,
        optional=False,
    ):
        QgsProcessingParameterNumber.__init__(
            self,
            name,
            description,
            QgsProcessingParameterNumber.Type.Double,
            default,
            optional,
            minValue,
        )
        self.setMetadata(
            {
                "widget_wrapper": "processing.algs.qgis.ui.InterpolationWidgets.PixelSizeWidgetWrapper"
            }
        )

        self.layersData = layersData
        self.extent = extent
        self.layers = []

    def clone(self):
        copy = ParameterPixelSize(
            self.name(),
            self.description(),
            self.layersData,
            self.extent,
            self.minimum(),
            self.defaultValue(),
            self.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional,
        )
        return copy


WIDGET, BASE = uic.loadUiType(os.path.join(pluginPath, "RasterResolutionWidget.ui"))


class PixelSizeWidget(BASE, WIDGET):
    def __init__(self):
        super().__init__(None)
        self.setupUi(self)
        self.context = dataobjects.createContext()

        self.extent: Optional[Union[QgsReferencedRectangle, QgsRectangle]] = (
            QgsRectangle()
        )
        self.layers = []

        self.mCellXSpinBox.setShowClearButton(False)
        self.mCellYSpinBox.setShowClearButton(False)
        self.mRowsSpinBox.setShowClearButton(False)
        self.mColumnsSpinBox.setShowClearButton(False)

        self.mCellYSpinBox.valueChanged.connect(self.mCellXSpinBox.setValue)
        self.mCellXSpinBox.valueChanged.connect(self.pixelSizeChanged)
        self.mRowsSpinBox.valueChanged.connect(self.rowsChanged)
        self.mColumnsSpinBox.valueChanged.connect(self.columnsChanged)

    def setLayers(self, layersData: str):
        self.extent = QgsRectangle()
        self.layers = []
        for row in layersData.split(";"):
            v = row.split("::~::")
            # need to keep a reference until interpolation is complete
            layer = QgsProcessingUtils.variantToSource(v[0], self.context)
            if layer:
                self.layers.append(layer)
                self.extent.combineExtentWith(layer.sourceExtent())

        self.pixelSizeChanged()

    def setExtent(self, extent: Optional[str]):
        if extent is not None:
            tokens = extent.split(" ")[0].split(",")
            ext = QgsRectangle(
                float(tokens[0]), float(tokens[2]), float(tokens[1]), float(tokens[3])
            )

            if len(tokens) > 1:
                self.extent = QgsReferencedRectangle(
                    ext, QgsCoordinateReferenceSystem(tokens[1][1:-1])
                )
            else:
                self.extent = ext

        self.pixelSizeChanged()

    def pixelSizeChanged(self):
        cell_size = self.mCellXSpinBox.value()
        if cell_size <= 0 or self.extent.isNull():
            return

        self.mCellYSpinBox.blockSignals(True)
        self.mCellYSpinBox.setValue(cell_size)
        self.mCellYSpinBox.blockSignals(False)
        rows = max(round(self.extent.height() / cell_size) + 1, 1)
        cols = max(round(self.extent.width() / cell_size) + 1, 1)
        self.mRowsSpinBox.blockSignals(True)
        self.mRowsSpinBox.setValue(rows)
        self.mRowsSpinBox.blockSignals(False)
        self.mColumnsSpinBox.blockSignals(True)
        self.mColumnsSpinBox.setValue(cols)
        self.mColumnsSpinBox.blockSignals(False)

    def rowsChanged(self):
        rows = self.mRowsSpinBox.value()
        if rows <= 0 or self.extent.isNull():
            return

        cell_size = self.extent.height() / rows
        if cell_size == 0:
            return

        cols = max(round(self.extent.width() / cell_size) + 1, 1)
        self.mColumnsSpinBox.blockSignals(True)
        self.mColumnsSpinBox.setValue(cols)
        self.mColumnsSpinBox.blockSignals(False)
        for w in [self.mCellXSpinBox, self.mCellYSpinBox]:
            w.blockSignals(True)
            w.setValue(cell_size)
            w.blockSignals(False)

    def columnsChanged(self):
        cols = self.mColumnsSpinBox.value()
        if cols < 2 or self.extent.isNull():
            return

        cell_size = self.extent.width() / (cols - 1)
        if cell_size == 0:
            return

        rows = max(round(self.extent.height() / cell_size), 1)
        self.mRowsSpinBox.blockSignals(True)
        self.mRowsSpinBox.setValue(rows)
        self.mRowsSpinBox.blockSignals(False)
        for w in [self.mCellXSpinBox, self.mCellYSpinBox]:
            w.blockSignals(True)
            w.setValue(cell_size)
            w.blockSignals(False)

    def setValue(self, value):
        try:
            numeric_value = float(value)
        except:
            return False

        self.mCellXSpinBox.setValue(numeric_value)
        self.mCellYSpinBox.setValue(numeric_value)
        return True

    def value(self):
        return self.mCellXSpinBox.value()


class PixelSizeWidgetWrapper(WidgetWrapper):
    def __init__(self, param, dialog, row=0, col=0, **kwargs):
        super().__init__(param, dialog, row, col, **kwargs)
        self.context = dataobjects.createContext()

    def _panel(self):
        return PixelSizeWidget()

    def createWidget(self):
        if self.dialogType == DIALOG_STANDARD:
            return self._panel()
        else:
            w = QgsDoubleSpinBox()
            w.setShowClearButton(False)
            w.setMinimum(0)
            w.setMaximum(99999999999)
            w.setDecimals(6)
            w.setToolTip(
                self.tr("Resolution of each pixel in output raster, in layer units")
            )
            return w

    def postInitialize(self, wrappers):
        if self.dialogType != DIALOG_STANDARD:
            return

        for wrapper in wrappers:
            if (
                wrapper.parameterDefinition().name()
                == self.parameterDefinition().layersData
            ):
                self.setLayers(wrapper.parameterValue())
                wrapper.widgetValueHasChanged.connect(self.layersChanged)
            elif (
                wrapper.parameterDefinition().name()
                == self.parameterDefinition().extent
            ):
                self.setExtent(wrapper.parameterValue())
                wrapper.widgetValueHasChanged.connect(self.extentChanged)

    def layersChanged(self, wrapper):
        self.setLayers(wrapper.parameterValue())

    def setLayers(self, layersData: str):
        self.widget.setLayers(layersData)

    def extentChanged(self, wrapper):
        self.setExtent(wrapper.parameterValue())

    def setExtent(self, extent: Optional[str]):
        self.widget.setExtent(extent)

    def setValue(self, value):
        return self.widget.setValue(value)

    def value(self):
        return self.widget.value()



"""
***************************************************************************
    g_extension_manage.py
    ------------------
    Date                 : May 2023
    Copyright            : (C) 2023 by Alister Hood
    Email                : alister.hood at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alister Hood"
__date__ = "May 2023"
__copyright__ = "(C) 2023, Alister Hood"


def processInputs(alg, parameters, context, feedback):
    pass



# -*- coding: utf-8 -*-

"""
***************************************************************************
    QtGui.py
    ---------------------
    Date                 : November 2015
    Copyright            : (C) 2015 by Matthias Kuhn
    Email                : matthias at opengis dot ch
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = 'Matthias Kuhn'
__date__ = 'November 2015'
__copyright__ = '(C) 2015, Matthias Kuhn'


from PyQt6.QtCore import QT_VERSION
from PyQt6.QtGui import *
def __qcolor_repr__(self: QColor):
    import math
    def _not_fuzzy_equal(a, b):
        return not math.isclose(a, b, abs_tol=0.00001)

    if not self.isValid():
        return '<QColor: invalid>'
    elif self.spec() == QColor.Spec.Rgb:
        if (_not_fuzzy_equal(self.red(), self.redF() * 255)
                or _not_fuzzy_equal(self.green(), self.greenF() * 255)
                or _not_fuzzy_equal(self.blue(), self.blueF() * 255)
                or _not_fuzzy_equal(self.alpha(), self.alphaF() * 255)):
            # represent using floats if required for full precision
            return f'<QColor: RGBA {self.redF() * 255}, {self.greenF() * 255}, {self.blueF() * 255}, {self.alphaF() * 255}>'
        return f'<QColor: RGBA {self.red()}, {self.green()}, {self.blue()}, {self.alpha()}>'
    elif self.spec() == QColor.Spec.Hsv:
        if (_not_fuzzy_equal(self.hsvHue(), self.hsvHueF() * 360)
                or _not_fuzzy_equal(self.hsvSaturation(),
                                    self.hsvSaturationF() * 255)
                or _not_fuzzy_equal(self.value(), self.valueF() * 255)
                or _not_fuzzy_equal(self.alpha(), self.alphaF() * 255)):
            # represent using floats if required for full precision
            return f'<QColor: HSVA {self.hsvHueF() * 360}, {self.hsvSaturationF() * 255}, {self.valueF() * 255}, {self.alphaF() * 255}>'
        return f'<QColor: HSVA {self.hsvHue()}, {self.hsvSaturation()}, {self.value()}, {self.alpha()}>'
    elif self.spec() == QColor.Spec.Cmyk:
        if (_not_fuzzy_equal(self.cyan(), self.cyanF() * 255)
                or _not_fuzzy_equal(self.magenta(), self.magentaF() * 255)
                or _not_fuzzy_equal(self.yellow(), self.yellowF() * 255)
                or _not_fuzzy_equal(self.black(), self.blackF() * 255)
                or _not_fuzzy_equal(self.alpha(), self.alphaF() * 255)):
            # represent using floats if required for full precision
            return f'<QColor: CMYKA {self.cyan()}, {self.magenta()}, {self.yellow()}, {self.black()}, {self.alpha()}>'
        return f'<QColor: CMYKA {self.cyan()}, {self.magenta()}, {self.yellow()}, {self.black()}, {self.alpha()}>'
    elif self.spec() == QColor.Spec.Hsl:
        if (_not_fuzzy_equal(self.hslHue(), self.hslHueF() * 360)
                or _not_fuzzy_equal(self.hslSaturation(),
                                    self.hslSaturationF() * 255)
                or _not_fuzzy_equal(self.lightness(), self.lightnessF() * 255)
                or _not_fuzzy_equal(self.alpha(), self.alphaF() * 255)):
            # represent using floats if required for full precision
            return f'<QColor: HSLA {self.hslHueF() * 360}, {self.hslSaturationF() * 255}, {self.lightnessF() * 255}, {self.alphaF() * 255}>'
        return f'<QColor: HSLA {self.hslHue()}, {self.hslSaturation()}, {self.lightness()}, {self.alpha()}>'
    elif self.spec() == QColor.Spec.ExtendedRgb:
        return f'<QColor: Extended RGBA {self.redF()}, {self.greenF()}, {self.blueF()}, {self.alphaF()}>'


# PyQt doesn't provide __repr__ for QColor, but it's highly desirable!
QColor.__repr__ = __qcolor_repr__

if 6 == 6:
  # patch back in Qt flags removed in PyQt
  QFileSystemModel.Options = lambda flags=0: QFileSystemModel.Option(flags)
  QGlyphRun.GlyphRunFlags = lambda flags=0: QGlyphRun.GlyphRunFlag(flags)
  QImageIOHandler.Transformations = lambda flags=0: QImageIOHandler.Transformation(flags)
  QPaintEngine.DirtyFlags = lambda flags=0: QPaintEngine.DirtyFlag(flags)
  QPaintEngine.PaintEngineFeatures = lambda flags=0: QPaintEngine.PaintEngineFeature(flags)
  QPainter.PixmapFragmentHints = lambda flags=0: QPainter.PixmapFragmentHint(flags)
  QPainter.RenderHints = lambda flags=0: QPainter.RenderHint(flags)
  QRawFont.LayoutFlags = lambda flags=0: QRawFont.LayoutFlag(flags)
  QSurfaceFormat.FormatOptions = lambda flags=0: QSurfaceFormat.FormatOption(flags)
  QTextFormat.PageBreakFlags = lambda flags=0: QTextFormat.PageBreakFlag(flags)
  QTextDocument.FindFlags = lambda flags=0: QTextDocument.FindFlag(flags)
  QTextDocument.MarkdownFeatures = lambda flags=0: QTextDocument.MarkdownFeature(flags)
  QTextItem.RenderFlags = lambda flags=0: QTextItem.RenderFlag(flags)
  QTextOption.Flags = lambda flags=0: QTextOption.Flag(flags)



"""
***************************************************************************
    grass_provider.py
    ---------------------
    Date                 : April 2014
    Copyright            : (C) 2014 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "April 2014"
__copyright__ = "(C) 2014, Victor Olaya"

import json

from processing.core.ProcessingConfig import ProcessingConfig, Setting
from qgis.core import (
    Qgis,
    QgsApplication,
    QgsMessageLog,
    QgsProcessingAlgorithm,
    QgsProcessingProvider,
    QgsRuntimeProfiler,
)
from qgis.PyQt.QtCore import QCoreApplication

from grassprovider.grass_algorithm import GrassAlgorithm
from grassprovider.grass_utils import GrassUtils


class GrassProvider(QgsProcessingProvider):
    def __init__(self):
        super().__init__()

    def load(self):
        with QgsRuntimeProfiler.profile("Grass Provider"):
            ProcessingConfig.settingIcons[self.name()] = self.icon()
            ProcessingConfig.addSetting(
                Setting(
                    self.name(),
                    GrassUtils.GRASS_LOG_COMMANDS,
                    self.tr("Log execution commands"),
                    False,
                )
            )
            ProcessingConfig.addSetting(
                Setting(
                    self.name(),
                    GrassUtils.GRASS_LOG_CONSOLE,
                    self.tr("Log console output"),
                    False,
                )
            )
            ProcessingConfig.addSetting(
                Setting(
                    self.name(),
                    GrassUtils.GRASS_HELP_URL,
                    self.tr("Location of GRASS docs"),
                    "",
                )
            )
            # Add settings for using r.external/v.external instead of r.in.gdal/v.in.ogr
            # but set them to False by default because the {r,v}.external implementations
            # have some bugs on windows + there are algorithms that can't be used with
            # external data (need a solid r.in.gdal/v.in.ogr).
            # For more info have a look at e.g. https://trac.osgeo.org/grass/ticket/3927
            ProcessingConfig.addSetting(
                Setting(
                    self.name(),
                    GrassUtils.GRASS_USE_REXTERNAL,
                    self.tr(
                        "For raster layers, use r.external (faster) instead of r.in.gdal"
                    ),
                    False,
                )
            )
            ProcessingConfig.addSetting(
                Setting(
                    self.name(),
                    GrassUtils.GRASS_USE_VEXTERNAL,
                    self.tr(
                        "For vector layers, use v.external (faster) instead of v.in.ogr"
                    ),
                    False,
                )
            )
            ProcessingConfig.readSettings()
            self.refreshAlgorithms()

        return True

    def unload(self):
        ProcessingConfig.removeSetting(GrassUtils.GRASS_LOG_COMMANDS)
        ProcessingConfig.removeSetting(GrassUtils.GRASS_LOG_CONSOLE)
        ProcessingConfig.removeSetting(GrassUtils.GRASS_HELP_URL)
        ProcessingConfig.removeSetting(GrassUtils.GRASS_USE_REXTERNAL)
        ProcessingConfig.removeSetting(GrassUtils.GRASS_USE_VEXTERNAL)

    def parse_algorithms(self) -> list[QgsProcessingAlgorithm]:
        """
        Parses all algorithm sources and returns a list of all GRASS
        algorithms.
        """
        algs = []
        for folder in GrassUtils.grassDescriptionFolders():
            if (folder / "algorithms.json").exists():
                # fast approach -- use aggregated JSON summary of algorithms
                with open(folder / "algorithms.json", encoding="utf8") as f_in:
                    algorithm_strings = f_in.read()

                algorithms_json = json.loads(algorithm_strings)
                for algorithm_json in algorithms_json:
                    try:
                        alg = GrassAlgorithm(
                            json_definition=algorithm_json, description_folder=folder
                        )
                        if alg.name().strip() != "":
                            algs.append(alg)
                        else:
                            QgsMessageLog.logMessage(
                                self.tr("Could not open GRASS algorithm: {0}").format(
                                    algorithm_json.get("name")
                                ),
                                self.tr("Processing"),
                                Qgis.MessageLevel.Critical,
                            )
                    except Exception as e:
                        QgsMessageLog.logMessage(
                            self.tr("Could not open GRASS algorithm: {0}\n{1}").format(
                                algorithm_json.get("name"), e
                            ),
                            self.tr("Processing"),
                            Qgis.MessageLevel.Critical,
                        )
            else:
                # slow approach - pass txt files one by one
                for descriptionFile in folder.glob("*.txt"):
                    try:
                        alg = GrassAlgorithm(description_file=descriptionFile)
                        if alg.name().strip() != "":
                            algs.append(alg)
                        else:
                            QgsMessageLog.logMessage(
                                self.tr("Could not open GRASS algorithm: {0}").format(
                                    descriptionFile
                                ),
                                self.tr("Processing"),
                                Qgis.MessageLevel.Critical,
                            )
                    except Exception as e:
                        QgsMessageLog.logMessage(
                            self.tr("Could not open GRASS algorithm: {0}\n{1}").format(
                                descriptionFile, e
                            ),
                            self.tr("Processing"),
                            Qgis.MessageLevel.Critical,
                        )
        return algs

    def loadAlgorithms(self):
        for a in self.parse_algorithms():
            self.addAlgorithm(a)

    def name(self):
        return "GRASS"

    def longName(self):
        version = GrassUtils.installedVersion()
        return f"GRASS ({version})" if version is not None else "GRASS"

    def id(self):
        return "grass"

    def helpId(self):
        return "grass7"

    def icon(self):
        return QgsApplication.getThemeIcon("/providerGrass.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("/providerGrass.svg")

    def versionInfo(self):
        return GrassUtils.installedVersion() or None

    def supportsNonFileBasedOutput(self):
        """
        GRASS Provider doesn't support non file based outputs
        """
        return False

    def supportedOutputVectorLayerExtensions(self):
        # We use the same extensions as QGIS because:
        # - QGIS is using OGR like GRASS
        # - There are very few chances that OGR version used in GRASS is
        # different from QGIS OGR version.
        return super().supportedOutputVectorLayerExtensions()

    def getSupportedOutputRasterFormatAndExtensions(self):
        return GrassUtils.getSupportedOutputRasterFormatAndExtensions()

    def canBeActivated(self):
        return not bool(GrassUtils.checkGrassIsInstalled())

    def tr(self, string: str):
        return QCoreApplication.translate(self.__class__.__name__, string)



# -*- coding: utf-8 -*-

"""
***************************************************************************
    __init__.py
    ---------------------
    Date                 : October 2014
    Copyright            : (C) 2014 by Alessandro Pasotti
    Email                : elpaso at itopen dot it
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = 'Alessandro Pasotti'
__date__ = 'October 2014'
__copyright__ = '(C) 2014, Alessandro Pasotti'

from qgis.PyQt import QtCore     # NOQA

from qgis._server import *  # NOQA

"""
This folder is completed using sipify.py script
It is not aimed to be manually edited
"""
# The following has been generated automatically from src/server/qgsaccesscontrol.h
try:
    QgsAccessControl.__overridden_methods__ = ['isFilterThreadSafe', 'filterFeatures', 'layerAttributes', 'clone']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsaccesscontrolfilter.h
try:
    QgsAccessControlFilter.__virtual_methods__ = ['layerFilterExpression', 'layerFilterSubsetString', 'layerPermissions', 'authorizedLayerAttributes', 'allowToEdit', 'cacheKey']
except (NameError, AttributeError):
    pass
try:
    QgsAccessControlFilter.LayerPermissions.__doc__ = """Describe the layer permission"""
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsbufferserverrequest.h
try:
    QgsBufferServerRequest.__overridden_methods__ = ['data']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsbufferserverresponse.h
try:
    QgsBufferServerResponse.__overridden_methods__ = ['setHeader', 'addHeader', 'removeHeader', 'header', 'fullHeaders', 'fullHeader', 'headers', 'headersSent', 'setStatusCode', 'statusCode', 'sendError', 'io', 'finish', 'flush', 'clear', 'data', 'truncate']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsconfigcache.h
try:
    QgsConfigCache.__attribute_docs__ = {'projectRemovedFromCache': 'Emitted whenever a project is removed from the cache.\n\n.. versionadded:: 3.38\n'}
    QgsConfigCache.initialize = staticmethod(QgsConfigCache.initialize)
    QgsConfigCache.instance = staticmethod(QgsConfigCache.instance)
    QgsConfigCache.__signal_arguments__ = {'projectRemovedFromCache': ['path: str']}
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsfcgiserverrequest.h
try:
    QgsFcgiServerRequest.__overridden_methods__ = ['data', 'header']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsfeaturefilter.h
try:
    QgsFeatureFilter.__overridden_methods__ = ['isFilterThreadSafe', 'filterFeatures', 'layerAttributes', 'clone']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsfeaturefilterprovidergroup.h
try:
    QgsFeatureFilterProviderGroup.__overridden_methods__ = ['isFilterThreadSafe', 'filterFeatures', 'layerAttributes', 'clone']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverapi.h
try:
    QgsServerApi.__virtual_methods__ = ['accept']
    QgsServerApi.__abstract_methods__ = ['executeRequest']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverapiutils.h
try:
    QgsServerApiUtils.parseBbox = staticmethod(QgsServerApiUtils.parseBbox)
    QgsServerApiUtils.temporalDimensions = staticmethod(QgsServerApiUtils.temporalDimensions)
    QgsServerApiUtils.parseTemporalDateInterval = staticmethod(QgsServerApiUtils.parseTemporalDateInterval)
    QgsServerApiUtils.parseTemporalDateTimeInterval = staticmethod(QgsServerApiUtils.parseTemporalDateTimeInterval)
    QgsServerApiUtils.fieldName = staticmethod(QgsServerApiUtils.fieldName)
    QgsServerApiUtils.temporalFilterExpression = staticmethod(QgsServerApiUtils.temporalFilterExpression)
    QgsServerApiUtils.temporalExtent = staticmethod(QgsServerApiUtils.temporalExtent)
    QgsServerApiUtils.parseCrs = staticmethod(QgsServerApiUtils.parseCrs)
    QgsServerApiUtils.sanitizedFieldValue = staticmethod(QgsServerApiUtils.sanitizedFieldValue)
    QgsServerApiUtils.publishedCrsList = staticmethod(QgsServerApiUtils.publishedCrsList)
    QgsServerApiUtils.crsToOgcUri = staticmethod(QgsServerApiUtils.crsToOgcUri)
    QgsServerApiUtils.appendMapParameter = staticmethod(QgsServerApiUtils.appendMapParameter)
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsservercachefilter.h
try:
    QgsServerCacheFilter.__virtual_methods__ = ['getCachedDocument', 'setCachedDocument', 'deleteCachedDocument', 'deleteCachedDocuments', 'getCachedImage', 'setCachedImage', 'deleteCachedImage', 'deleteCachedImages']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverexception.h
try:
    QgsServerException.__virtual_methods__ = ['formatResponse']
except (NameError, AttributeError):
    pass
try:
    QgsOgcServiceException.__overridden_methods__ = ['formatResponse']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverfeatureid.h
try:
    QgsServerFeatureId.getServerFid = staticmethod(QgsServerFeatureId.getServerFid)
    QgsServerFeatureId.updateFeatureRequestFromServerFids = staticmethod(QgsServerFeatureId.updateFeatureRequestFromServerFids)
    QgsServerFeatureId.getExpressionFromServerFid = staticmethod(QgsServerFeatureId.getExpressionFromServerFid)
    QgsServerFeatureId.pkSeparator = staticmethod(QgsServerFeatureId.pkSeparator)
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverfilter.h
try:
    QgsServerFilter.__virtual_methods__ = ['requestReady', 'responseComplete', 'sendResponse', 'onRequestReady', 'onProjectReady', 'onResponseComplete', 'onSendResponse']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverinterface.h
try:
    QgsServerInterface.__abstract_methods__ = ['capabilitiesCache', 'requestHandler', 'registerFilter', 'setFilters', 'filters', 'registerAccessControl', 'accessControls', 'registerServerCache', 'cacheManager', 'getEnv', 'configFilePath', 'setConfigFilePath', 'removeConfigCacheEntry', 'serviceRegistry', 'reloadSettings']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverlogger.h
try:
    QgsServerLogger.instance = staticmethod(QgsServerLogger.instance)
    QgsServerLogger.__overridden_methods__ = ['logMessage']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverogcapi.h
QgsServerOgcApi.alternate = QgsServerOgcApi.Rel.alternate
QgsServerOgcApi.describedBy = QgsServerOgcApi.Rel.describedBy
QgsServerOgcApi.collection = QgsServerOgcApi.Rel.collection
QgsServerOgcApi.item = QgsServerOgcApi.Rel.item
QgsServerOgcApi.self = QgsServerOgcApi.Rel.self
QgsServerOgcApi.service_desc = QgsServerOgcApi.Rel.service_desc
QgsServerOgcApi.service_doc = QgsServerOgcApi.Rel.service_doc
QgsServerOgcApi.prev = QgsServerOgcApi.Rel.prev
QgsServerOgcApi.next = QgsServerOgcApi.Rel.next
QgsServerOgcApi.license = QgsServerOgcApi.Rel.license
QgsServerOgcApi.items = QgsServerOgcApi.Rel.items
QgsServerOgcApi.conformance = QgsServerOgcApi.Rel.conformance
QgsServerOgcApi.data = QgsServerOgcApi.Rel.data
QgsServerOgcApi.Rel.baseClass = QgsServerOgcApi
QgsServerOgcApi.GEOJSON = QgsServerOgcApi.ContentType.GEOJSON
QgsServerOgcApi.OPENAPI3 = QgsServerOgcApi.ContentType.OPENAPI3
QgsServerOgcApi.JSON = QgsServerOgcApi.ContentType.JSON
QgsServerOgcApi.HTML = QgsServerOgcApi.ContentType.HTML
QgsServerOgcApi.XML = QgsServerOgcApi.ContentType.XML
QgsServerOgcApi.FLATGEOBUF = QgsServerOgcApi.ContentType.FLATGEOBUF
QgsServerOgcApi.ContentType.baseClass = QgsServerOgcApi
# monkey patching scoped based enum
QgsServerOgcApi.Profile.NONE.__doc__ = "No profile"
QgsServerOgcApi.Profile.RFC7946.__doc__ = "GeoJSON profile according to RFC7946"
QgsServerOgcApi.Profile.__doc__ = """JSON profile

* ``NONE``: No profile
* ``RFC7946``: GeoJSON profile according to RFC7946

"""
# --
QgsServerOgcApi.Profile.baseClass = QgsServerOgcApi
try:
    QgsServerOgcApi.sanitizeUrl = staticmethod(QgsServerOgcApi.sanitizeUrl)
    QgsServerOgcApi.relToString = staticmethod(QgsServerOgcApi.relToString)
    QgsServerOgcApi.contentTypeToString = staticmethod(QgsServerOgcApi.contentTypeToString)
    QgsServerOgcApi.contentTypeToStdString = staticmethod(QgsServerOgcApi.contentTypeToStdString)
    QgsServerOgcApi.contentTypeToExtension = staticmethod(QgsServerOgcApi.contentTypeToExtension)
    QgsServerOgcApi.contenTypeFromExtension = staticmethod(QgsServerOgcApi.contenTypeFromExtension)
    QgsServerOgcApi.contentTypeFromExtension = staticmethod(QgsServerOgcApi.contentTypeFromExtension)
    QgsServerOgcApi.mimeType = staticmethod(QgsServerOgcApi.mimeType)
    QgsServerOgcApi.__overridden_methods__ = ['name', 'description', 'version', 'rootPath', 'executeRequest']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverogcapihandler.h
try:
    QgsServerOgcApiHandler.parentLink = staticmethod(QgsServerOgcApiHandler.parentLink)
    QgsServerOgcApiHandler.layerFromCollectionId = staticmethod(QgsServerOgcApiHandler.layerFromCollectionId)
    QgsServerOgcApiHandler.__virtual_methods__ = ['parameters', 'tags', 'defaultContentType', 'handleRequest', 'values']
    QgsServerOgcApiHandler.__abstract_methods__ = ['path', 'operationId', 'summary', 'description', 'linkTitle', 'linkType']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverparameters.h
QgsServerParameter.UNKNOWN = QgsServerParameter.Name.UNKNOWN
QgsServerParameter.SERVICE = QgsServerParameter.Name.SERVICE
QgsServerParameter.VERSION_SERVICE = QgsServerParameter.Name.VERSION_SERVICE
QgsServerParameter.REQUEST = QgsServerParameter.Name.REQUEST
QgsServerParameter.MAP = QgsServerParameter.Name.MAP
QgsServerParameter.FILE_NAME = QgsServerParameter.Name.FILE_NAME
QgsServerParameter.Name.baseClass = QgsServerParameter
try:
    QgsServerParameterDefinition.raiseError = staticmethod(QgsServerParameterDefinition.raiseError)
    QgsServerParameterDefinition.__virtual_methods__ = ['isValid']
except (NameError, AttributeError):
    pass
try:
    QgsServerParameter.name = staticmethod(QgsServerParameter.name)
except (NameError, AttributeError):
    pass
try:
    QgsServerParameters.__virtual_methods__ = ['request', 'version', 'loadParameter']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverprojectutils.h
try:
    QgsServerProjectUtils.owsServiceCapabilities = staticmethod(QgsServerProjectUtils.owsServiceCapabilities)
    QgsServerProjectUtils.owsServiceTitle = staticmethod(QgsServerProjectUtils.owsServiceTitle)
    QgsServerProjectUtils.owsServiceAbstract = staticmethod(QgsServerProjectUtils.owsServiceAbstract)
    QgsServerProjectUtils.owsServiceKeywords = staticmethod(QgsServerProjectUtils.owsServiceKeywords)
    QgsServerProjectUtils.owsServiceOnlineResource = staticmethod(QgsServerProjectUtils.owsServiceOnlineResource)
    QgsServerProjectUtils.owsServiceContactOrganization = staticmethod(QgsServerProjectUtils.owsServiceContactOrganization)
    QgsServerProjectUtils.owsServiceContactPosition = staticmethod(QgsServerProjectUtils.owsServiceContactPosition)
    QgsServerProjectUtils.owsServiceContactPerson = staticmethod(QgsServerProjectUtils.owsServiceContactPerson)
    QgsServerProjectUtils.owsServiceContactMail = staticmethod(QgsServerProjectUtils.owsServiceContactMail)
    QgsServerProjectUtils.owsServiceContactPhone = staticmethod(QgsServerProjectUtils.owsServiceContactPhone)
    QgsServerProjectUtils.owsServiceFees = staticmethod(QgsServerProjectUtils.owsServiceFees)
    QgsServerProjectUtils.owsServiceAccessConstraints = staticmethod(QgsServerProjectUtils.owsServiceAccessConstraints)
    QgsServerProjectUtils.wmsMaxWidth = staticmethod(QgsServerProjectUtils.wmsMaxWidth)
    QgsServerProjectUtils.wmsMaxHeight = staticmethod(QgsServerProjectUtils.wmsMaxHeight)
    QgsServerProjectUtils.wmsImageQuality = staticmethod(QgsServerProjectUtils.wmsImageQuality)
    QgsServerProjectUtils.wmsTileBuffer = staticmethod(QgsServerProjectUtils.wmsTileBuffer)
    QgsServerProjectUtils.wmsRenderMapTiles = staticmethod(QgsServerProjectUtils.wmsRenderMapTiles)
    QgsServerProjectUtils.wmsMaxAtlasFeatures = staticmethod(QgsServerProjectUtils.wmsMaxAtlasFeatures)
    QgsServerProjectUtils.wmsDefaultMapUnitsPerMm = staticmethod(QgsServerProjectUtils.wmsDefaultMapUnitsPerMm)
    QgsServerProjectUtils.wmsUseLayerIds = staticmethod(QgsServerProjectUtils.wmsUseLayerIds)
    QgsServerProjectUtils.wmsInfoFormatSia2045 = staticmethod(QgsServerProjectUtils.wmsInfoFormatSia2045)
    QgsServerProjectUtils.wmsFeatureInfoAddWktGeometry = staticmethod(QgsServerProjectUtils.wmsFeatureInfoAddWktGeometry)
    QgsServerProjectUtils.wmsFeatureInfoUseAttributeFormSettings = staticmethod(QgsServerProjectUtils.wmsFeatureInfoUseAttributeFormSettings)
    QgsServerProjectUtils.wmsHTMLFeatureInfoUseOnlyMaptip = staticmethod(QgsServerProjectUtils.wmsHTMLFeatureInfoUseOnlyMaptip)
    QgsServerProjectUtils.wmsFeatureInfoSegmentizeWktGeometry = staticmethod(QgsServerProjectUtils.wmsFeatureInfoSegmentizeWktGeometry)
    QgsServerProjectUtils.wmsAddLegendGroupsLegendGraphic = staticmethod(QgsServerProjectUtils.wmsAddLegendGroupsLegendGraphic)
    QgsServerProjectUtils.wmsSkipNameForGroup = staticmethod(QgsServerProjectUtils.wmsSkipNameForGroup)
    QgsServerProjectUtils.wmsFeatureInfoPrecision = staticmethod(QgsServerProjectUtils.wmsFeatureInfoPrecision)
    QgsServerProjectUtils.wmsFeatureInfoDocumentElement = staticmethod(QgsServerProjectUtils.wmsFeatureInfoDocumentElement)
    QgsServerProjectUtils.wmsFeatureInfoDocumentElementNs = staticmethod(QgsServerProjectUtils.wmsFeatureInfoDocumentElementNs)
    QgsServerProjectUtils.wmsFeatureInfoSchema = staticmethod(QgsServerProjectUtils.wmsFeatureInfoSchema)
    QgsServerProjectUtils.wmsFeatureInfoLayerAliasMap = staticmethod(QgsServerProjectUtils.wmsFeatureInfoLayerAliasMap)
    QgsServerProjectUtils.wmsInspireActivate = staticmethod(QgsServerProjectUtils.wmsInspireActivate)
    QgsServerProjectUtils.wmsInspireLanguage = staticmethod(QgsServerProjectUtils.wmsInspireLanguage)
    QgsServerProjectUtils.wmsInspireMetadataUrl = staticmethod(QgsServerProjectUtils.wmsInspireMetadataUrl)
    QgsServerProjectUtils.wmsInspireMetadataUrlType = staticmethod(QgsServerProjectUtils.wmsInspireMetadataUrlType)
    QgsServerProjectUtils.wmsInspireTemporalReference = staticmethod(QgsServerProjectUtils.wmsInspireTemporalReference)
    QgsServerProjectUtils.wmsInspireMetadataDate = staticmethod(QgsServerProjectUtils.wmsInspireMetadataDate)
    QgsServerProjectUtils.wmsRestrictedComposers = staticmethod(QgsServerProjectUtils.wmsRestrictedComposers)
    QgsServerProjectUtils.wmsServiceUrl = staticmethod(QgsServerProjectUtils.wmsServiceUrl)
    QgsServerProjectUtils.wmsRootName = staticmethod(QgsServerProjectUtils.wmsRootName)
    QgsServerProjectUtils.wmsRestrictedLayers = staticmethod(QgsServerProjectUtils.wmsRestrictedLayers)
    QgsServerProjectUtils.wmsOutputCrsList = staticmethod(QgsServerProjectUtils.wmsOutputCrsList)
    QgsServerProjectUtils.wmsOutputCrsListAsOgcUrn = staticmethod(QgsServerProjectUtils.wmsOutputCrsListAsOgcUrn)
    QgsServerProjectUtils.wmsExtent = staticmethod(QgsServerProjectUtils.wmsExtent)
    QgsServerProjectUtils.wfsServiceUrl = staticmethod(QgsServerProjectUtils.wfsServiceUrl)
    QgsServerProjectUtils.wfsLayerIds = staticmethod(QgsServerProjectUtils.wfsLayerIds)
    QgsServerProjectUtils.wfsLayerPrecision = staticmethod(QgsServerProjectUtils.wfsLayerPrecision)
    QgsServerProjectUtils.wfstUpdateLayerIds = staticmethod(QgsServerProjectUtils.wfstUpdateLayerIds)
    QgsServerProjectUtils.wfstInsertLayerIds = staticmethod(QgsServerProjectUtils.wfstInsertLayerIds)
    QgsServerProjectUtils.wfstDeleteLayerIds = staticmethod(QgsServerProjectUtils.wfstDeleteLayerIds)
    QgsServerProjectUtils.wcsServiceUrl = staticmethod(QgsServerProjectUtils.wcsServiceUrl)
    QgsServerProjectUtils.wcsLayerIds = staticmethod(QgsServerProjectUtils.wcsLayerIds)
    QgsServerProjectUtils.wmtsServiceUrl = staticmethod(QgsServerProjectUtils.wmtsServiceUrl)
    QgsServerProjectUtils.serviceUrl = staticmethod(QgsServerProjectUtils.serviceUrl)
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverquerystringparameter.h
# monkey patching scoped based enum
QgsServerQueryStringParameter.Type.String.__doc__ = "Parameter is a string"
QgsServerQueryStringParameter.Type.Integer.__doc__ = "Parameter is an integer"
QgsServerQueryStringParameter.Type.Double.__doc__ = "Parameter is a double"
QgsServerQueryStringParameter.Type.Boolean.__doc__ = "Parameter is a boolean"
QgsServerQueryStringParameter.Type.List.__doc__ = "Parameter is a (comma separated) list of strings, the handler will perform any further required conversion of the list values"
QgsServerQueryStringParameter.Type.__doc__ = """The Type enum represents the parameter type

* ``String``: Parameter is a string
* ``Integer``: Parameter is an integer
* ``Double``: Parameter is a double
* ``Boolean``: Parameter is a boolean
* ``List``: Parameter is a (comma separated) list of strings, the handler will perform any further required conversion of the list values

"""
# --
QgsServerQueryStringParameter.Type.baseClass = QgsServerQueryStringParameter
try:
    QgsServerQueryStringParameter.typeName = staticmethod(QgsServerQueryStringParameter.typeName)
    QgsServerQueryStringParameter.__virtual_methods__ = ['value']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverrequest.h
QgsServerRequest.HeadMethod = QgsServerRequest.Method.HeadMethod
QgsServerRequest.PutMethod = QgsServerRequest.Method.PutMethod
QgsServerRequest.GetMethod = QgsServerRequest.Method.GetMethod
QgsServerRequest.PostMethod = QgsServerRequest.Method.PostMethod
QgsServerRequest.DeleteMethod = QgsServerRequest.Method.DeleteMethod
QgsServerRequest.PatchMethod = QgsServerRequest.Method.PatchMethod
QgsServerRequest.OptionsMethod = QgsServerRequest.Method.OptionsMethod
QgsServerRequest.Method.baseClass = QgsServerRequest
QgsServerRequest.HOST = QgsServerRequest.RequestHeader.HOST
QgsServerRequest.FORWARDED = QgsServerRequest.RequestHeader.FORWARDED
QgsServerRequest.X_FORWARDED_FOR = QgsServerRequest.RequestHeader.X_FORWARDED_FOR
QgsServerRequest.X_FORWARDED_HOST = QgsServerRequest.RequestHeader.X_FORWARDED_HOST
QgsServerRequest.X_FORWARDED_PROTO = QgsServerRequest.RequestHeader.X_FORWARDED_PROTO
QgsServerRequest.X_QGIS_SERVICE_URL = QgsServerRequest.RequestHeader.X_QGIS_SERVICE_URL
QgsServerRequest.X_QGIS_WMS_SERVICE_URL = QgsServerRequest.RequestHeader.X_QGIS_WMS_SERVICE_URL
QgsServerRequest.X_QGIS_WFS_SERVICE_URL = QgsServerRequest.RequestHeader.X_QGIS_WFS_SERVICE_URL
QgsServerRequest.X_QGIS_WCS_SERVICE_URL = QgsServerRequest.RequestHeader.X_QGIS_WCS_SERVICE_URL
QgsServerRequest.X_QGIS_WMTS_SERVICE_URL = QgsServerRequest.RequestHeader.X_QGIS_WMTS_SERVICE_URL
QgsServerRequest.ACCEPT = QgsServerRequest.RequestHeader.ACCEPT
QgsServerRequest.USER_AGENT = QgsServerRequest.RequestHeader.USER_AGENT
QgsServerRequest.AUTHORIZATION = QgsServerRequest.RequestHeader.AUTHORIZATION
QgsServerRequest.RequestHeader.baseClass = QgsServerRequest
try:
    QgsServerRequest.methodToString = staticmethod(QgsServerRequest.methodToString)
    QgsServerRequest.__virtual_methods__ = ['setParameter', 'removeParameter', 'header', 'data', 'setUrl']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverresponse.h
try:
    QgsServerResponse.__virtual_methods__ = ['write', 'finish', 'flush', 'feedback']
    QgsServerResponse.__abstract_methods__ = ['setHeader', 'addHeader', 'removeHeader', 'header', 'fullHeader', 'headers', 'fullHeaders', 'headersSent', 'setStatusCode', 'statusCode', 'sendError', 'io', 'clear', 'data', 'truncate']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserversettings.h
QgsServerSettingsEnv.DEFAULT_VALUE = QgsServerSettingsEnv.Source.DEFAULT_VALUE
QgsServerSettingsEnv.ENVIRONMENT_VARIABLE = QgsServerSettingsEnv.Source.ENVIRONMENT_VARIABLE
QgsServerSettingsEnv.INI_FILE = QgsServerSettingsEnv.Source.INI_FILE
QgsServerSettingsEnv.Source.baseClass = QgsServerSettingsEnv
QgsServerSettingsEnv.QGIS_OPTIONS_PATH = QgsServerSettingsEnv.EnvVar.QGIS_OPTIONS_PATH
QgsServerSettingsEnv.QGIS_SERVER_PARALLEL_RENDERING = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_PARALLEL_RENDERING
QgsServerSettingsEnv.QGIS_SERVER_MAX_THREADS = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_MAX_THREADS
QgsServerSettingsEnv.QGIS_SERVER_LOG_LEVEL = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_LOG_LEVEL
QgsServerSettingsEnv.QGIS_SERVER_LOG_FILE = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_LOG_FILE
QgsServerSettingsEnv.QGIS_SERVER_LOG_STDERR = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_LOG_STDERR
QgsServerSettingsEnv.QGIS_PROJECT_FILE = QgsServerSettingsEnv.EnvVar.QGIS_PROJECT_FILE
QgsServerSettingsEnv.QGIS_SERVER_IGNORE_BAD_LAYERS = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_IGNORE_BAD_LAYERS
QgsServerSettingsEnv.QGIS_SERVER_IGNORE_RENDERING_ERRORS = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_IGNORE_RENDERING_ERRORS
QgsServerSettingsEnv.QGIS_SERVER_RETRY_BAD_LAYERS = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_RETRY_BAD_LAYERS
QgsServerSettingsEnv.QGIS_SERVER_CACHE_DIRECTORY = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_CACHE_DIRECTORY
QgsServerSettingsEnv.QGIS_SERVER_CACHE_SIZE = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_CACHE_SIZE
QgsServerSettingsEnv.QGIS_SERVER_SHOW_GROUP_SEPARATOR = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_SHOW_GROUP_SEPARATOR
QgsServerSettingsEnv.QGIS_SERVER_OVERRIDE_SYSTEM_LOCALE = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_OVERRIDE_SYSTEM_LOCALE
QgsServerSettingsEnv.QGIS_SERVER_WMS_MAX_HEIGHT = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_WMS_MAX_HEIGHT
QgsServerSettingsEnv.QGIS_SERVER_WMS_MAX_WIDTH = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_WMS_MAX_WIDTH
QgsServerSettingsEnv.QGIS_SERVER_API_RESOURCES_DIRECTORY = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_API_RESOURCES_DIRECTORY
QgsServerSettingsEnv.QGIS_SERVER_API_WFS3_MAX_LIMIT = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_API_WFS3_MAX_LIMIT
QgsServerSettingsEnv.QGIS_SERVER_API_WFS3_ROOT_PATH = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_API_WFS3_ROOT_PATH
QgsServerSettingsEnv.QGIS_SERVER_TRUST_LAYER_METADATA = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_TRUST_LAYER_METADATA
QgsServerSettingsEnv.QGIS_SERVER_FORCE_READONLY_LAYERS = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_FORCE_READONLY_LAYERS
QgsServerSettingsEnv.QGIS_SERVER_DISABLE_GETPRINT = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_DISABLE_GETPRINT
QgsServerSettingsEnv.QGIS_SERVER_LANDING_PAGE_PROJECTS_DIRECTORIES = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_LANDING_PAGE_PROJECTS_DIRECTORIES
QgsServerSettingsEnv.QGIS_SERVER_LANDING_PAGE_PROJECTS_PG_CONNECTIONS = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_LANDING_PAGE_PROJECTS_PG_CONNECTIONS
QgsServerSettingsEnv.QGIS_SERVER_LOG_PROFILE = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_LOG_PROFILE
QgsServerSettingsEnv.QGIS_SERVER_SERVICE_URL = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_SERVICE_URL
QgsServerSettingsEnv.QGIS_SERVER_WMS_SERVICE_URL = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_WMS_SERVICE_URL
QgsServerSettingsEnv.QGIS_SERVER_WFS_SERVICE_URL = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_WFS_SERVICE_URL
QgsServerSettingsEnv.QGIS_SERVER_WCS_SERVICE_URL = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_WCS_SERVICE_URL
QgsServerSettingsEnv.QGIS_SERVER_WMTS_SERVICE_URL = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_WMTS_SERVICE_URL
QgsServerSettingsEnv.QGIS_SERVER_LANDING_PAGE_PREFIX = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_LANDING_PAGE_PREFIX
QgsServerSettingsEnv.QGIS_SERVER_PROJECT_CACHE_CHECK_INTERVAL = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_PROJECT_CACHE_CHECK_INTERVAL
QgsServerSettingsEnv.QGIS_SERVER_PROJECT_CACHE_STRATEGY = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_PROJECT_CACHE_STRATEGY
QgsServerSettingsEnv.QGIS_SERVER_PROJECT_CACHE_SIZE = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_PROJECT_CACHE_SIZE
QgsServerSettingsEnv.QGIS_SERVER_ALLOWED_EXTRA_SQL_TOKENS = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_ALLOWED_EXTRA_SQL_TOKENS
QgsServerSettingsEnv.QGIS_SERVER_APPLICATION_NAME = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_APPLICATION_NAME
QgsServerSettingsEnv.QGIS_SERVER_CAPABILITIES_CACHE_SIZE = QgsServerSettingsEnv.EnvVar.QGIS_SERVER_CAPABILITIES_CACHE_SIZE
QgsServerSettingsEnv.EnvVar.baseClass = QgsServerSettingsEnv
try:
    QgsServerSettings.name = staticmethod(QgsServerSettings.name)
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsserverstatichandler.h
try:
    QgsServerStaticHandler.__overridden_methods__ = ['handleRequest', 'path', 'operationId', 'summary', 'description', 'linkTitle', 'linkType']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsservice.h
try:
    QgsService.__abstract_methods__ = ['name', 'version', 'executeRequest']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsservicemodule.h
try:
    QgsServiceModule.__abstract_methods__ = ['registerSelf']
except (NameError, AttributeError):
    pass
# The following has been generated automatically from src/server/qgsstorebadlayerinfo.h
try:
    QgsStoreBadLayerInfo.__overridden_methods__ = ['handleBadLayers']
except (NameError, AttributeError):
    pass




"""
***************************************************************************
    extractprojection.py
    ---------------------
    Date                 : September 2013
    Copyright            : (C) 2013 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "September 2013"
__copyright__ = "(C) 2013, Alexander Bruy"

import os

from osgeo import gdal, osr
from qgis.core import (
    QgsProcessingException,
    QgsProcessingOutputFile,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterRasterLayer,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]

gdal.UseExceptions()
osr.UseExceptions()


class ExtractProjection(GdalAlgorithm):
    INPUT = "INPUT"
    PRJ_FILE_CREATE = "PRJ_FILE_CREATE"
    WORLD_FILE = "WORLD_FILE"
    PRJ_FILE = "PRJ_FILE"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Input file"))
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.PRJ_FILE_CREATE, self.tr("Create also .prj file"), False
            )
        )
        self.addOutput(QgsProcessingOutputFile(self.WORLD_FILE, self.tr("World file")))
        self.addOutput(
            QgsProcessingOutputFile(self.PRJ_FILE, self.tr("ESRI Shapefile prj file"))
        )

    def name(self):
        return "extractprojection"

    def displayName(self):
        return self.tr("Extract projection")

    def icon(self):
        return QIcon(
            os.path.join(pluginPath, "images", "gdaltools", "projection-export.png")
        )

    def group(self):
        return self.tr("Raster projections")

    def groupId(self):
        return "rasterprojections"

    def commandName(self):
        return "extractprojection"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        return [self.commandName()]

    def processAlgorithm(self, parameters, context, feedback):
        createPrj = self.parameterAsBoolean(parameters, self.PRJ_FILE_CREATE, context)
        raster = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        if raster.dataProvider().name() != "gdal":
            raise QgsProcessingException(
                self.tr("This algorithm can only be used with GDAL raster layers")
            )
        rasterPath = raster.source()
        rasterDS = gdal.Open(rasterPath, gdal.GA_ReadOnly)
        geotransform = rasterDS.GetGeoTransform()
        inputcrs = raster.crs()
        crs = inputcrs.toWkt()
        raster = None
        rasterDS = None

        inFileName = os.path.splitext(str(rasterPath))
        outFileName = inFileName[0]
        # this is not a good idea as it won't work with an extension like .jpeg
        # outFileExt = '.' + inFileName[1][1:4:2] + 'w'
        if len(inFileName[1]) < 4:
            outFileExt = ".wld"
        else:
            outFileExt = inFileName[1][0:2] + inFileName[1][-1] + "w"

        results = {}
        if crs != "" and createPrj:
            tmp = osr.SpatialReference()
            tmp.ImportFromWkt(crs)
            tmp.MorphToESRI()
            crs = tmp.ExportToWkt()
            tmp = None

            with open(outFileName + ".prj", "w", encoding="utf-8") as prj:
                prj.write(crs)
            results[self.PRJ_FILE] = outFileName + ".prj"
        else:
            results[self.PRJ_FILE] = None

        with open(outFileName + outFileExt, "w") as wld:
            wld.write("%0.8f\n" % geotransform[1])
            wld.write("%0.8f\n" % geotransform[4])
            wld.write("%0.8f\n" % geotransform[2])
            wld.write("%0.8f\n" % geotransform[5])
            wld.write(
                "%0.8f\n"
                % (geotransform[0] + 0.5 * geotransform[1] + 0.5 * geotransform[2])
            )
            wld.write(
                "%0.8f\n"
                % (geotransform[3] + 0.5 * geotransform[4] + 0.5 * geotransform[5])
            )
        results[self.WORLD_FILE] = outFileName + outFileExt

        return results



"""
***************************************************************************
    ParametersTest
    ---------------------
    Date                 : May 2021
    Copyright            : (C) 2021 by René-Luc DHONT
    Email                : rldhont at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "René-Luc DHONT"
__date__ = "May 2021"
__copyright__ = "(C) 2021, René-Luc DHONT"

import os
import shutil
import unittest

from qgis.core import (
    QgsProcessing,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterFile,
    QgsProcessingParameterNumber,
)
from qgis.testing import QgisTestCase, start_app

from processing.core.parameters import getParameterFromString

testDataPath = os.path.join(os.path.dirname(__file__), "testdata")

start_app()


class ParametersTest(QgisTestCase):
    @classmethod
    def setUpClass(cls):
        cls.cleanup_paths = []

    @classmethod
    def tearDownClass(cls):
        for path in cls.cleanup_paths:
            shutil.rmtree(path)

    def testParameterStringDesc(self):
        desc = "QgsProcessingParameterString|in_string|Input String"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "string")
        self.assertEqual(param.name(), "in_string")
        self.assertEqual(param.description(), "Input String")
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterString|in_string|Input String|default value"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "string")
        self.assertEqual(param.name(), "in_string")
        self.assertEqual(param.description(), "Input String")
        self.assertEqual(param.defaultValue(), "default value")
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterString|in_string|Input String|default value|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "string")
        self.assertEqual(param.name(), "in_string")
        self.assertEqual(param.description(), "Input String")
        self.assertEqual(param.defaultValue(), "default value")
        self.assertTrue(param.multiLine())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterString|in_string|Input String||False|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "string")
        self.assertEqual(param.name(), "in_string")
        self.assertEqual(param.description(), "Input String")
        self.assertEqual(param.defaultValue(), "")
        self.assertFalse(param.multiLine())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

    def testParameterNumberDesc(self):
        desc = "QgsProcessingParameterNumber|in_number|Input Number"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "number")
        self.assertEqual(param.name(), "in_number")
        self.assertEqual(param.description(), "Input Number")
        self.assertEqual(param.dataType(), QgsProcessingParameterNumber.Type.Integer)
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterNumber|in_number|Input Number|QgsProcessingParameterNumber.Double"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "number")
        self.assertEqual(param.name(), "in_number")
        self.assertEqual(param.description(), "Input Number")
        self.assertEqual(param.dataType(), QgsProcessingParameterNumber.Type.Double)
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterNumber|in_number|Input Number|QgsProcessingParameterNumber.Integer|10"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "number")
        self.assertEqual(param.name(), "in_number")
        self.assertEqual(param.description(), "Input Number")
        self.assertEqual(param.dataType(), QgsProcessingParameterNumber.Type.Integer)
        self.assertEqual(param.defaultValue(), 10)
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterNumber|in_number|Input Number|QgsProcessingParameterNumber.Integer|None|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "number")
        self.assertEqual(param.name(), "in_number")
        self.assertEqual(param.description(), "Input Number")
        self.assertEqual(param.dataType(), QgsProcessingParameterNumber.Type.Integer)
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterNumber|in_number|Input Number|QgsProcessingParameterNumber.Integer|10|False|0"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "number")
        self.assertEqual(param.name(), "in_number")
        self.assertEqual(param.description(), "Input Number")
        self.assertEqual(param.dataType(), QgsProcessingParameterNumber.Type.Integer)
        self.assertEqual(param.defaultValue(), 10)
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertEqual(param.minimum(), 0)

        desc = "QgsProcessingParameterNumber|in_number|Input Number|QgsProcessingParameterNumber.Integer|10|False|0|20"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "number")
        self.assertEqual(param.name(), "in_number")
        self.assertEqual(param.description(), "Input Number")
        self.assertEqual(param.dataType(), QgsProcessingParameterNumber.Type.Integer)
        self.assertEqual(param.defaultValue(), 10)
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertEqual(param.minimum(), 0)
        self.assertEqual(param.maximum(), 20)

    def testParameterBooleanDesc(self):
        desc = "QgsProcessingParameterBoolean|in_bool|Input Boolean"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "boolean")
        self.assertEqual(param.name(), "in_bool")
        self.assertEqual(param.description(), "Input Boolean")
        self.assertFalse(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterBoolean|in_bool|Input Boolean|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "boolean")
        self.assertEqual(param.name(), "in_bool")
        self.assertEqual(param.description(), "Input Boolean")
        self.assertTrue(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterBoolean|in_bool|Input Boolean|False|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "boolean")
        self.assertEqual(param.name(), "in_bool")
        self.assertEqual(param.description(), "Input Boolean")
        self.assertFalse(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

    def testParameterCrsDesc(self):
        desc = "QgsProcessingParameterCrs|in_crs|Input CRS"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "crs")
        self.assertEqual(param.name(), "in_crs")
        self.assertEqual(param.description(), "Input CRS")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterCrs|in_crs|Input CRS|EPSG:2154"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "crs")
        self.assertEqual(param.name(), "in_crs")
        self.assertEqual(param.description(), "Input CRS")
        self.assertEqual(param.defaultValue(), "EPSG:2154")
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterCrs|in_crs|Input CRS|None|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "crs")
        self.assertEqual(param.name(), "in_crs")
        self.assertEqual(param.description(), "Input CRS")
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

    def testParameterExtentDesc(self):
        desc = "QgsProcessingParameterExtent|in_extent|Input Extent"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "extent")
        self.assertEqual(param.name(), "in_extent")
        self.assertEqual(param.description(), "Input Extent")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterExtent|in_extent|Input Extent|None|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "extent")
        self.assertEqual(param.name(), "in_extent")
        self.assertEqual(param.description(), "Input Extent")
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

    def testParameterFileDesc(self):
        desc = "QgsProcessingParameterFile|in_file|Input File"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "file")
        self.assertEqual(param.name(), "in_file")
        self.assertEqual(param.description(), "Input File")
        self.assertEqual(param.behavior(), QgsProcessingParameterFile.Behavior.File)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFile|in_folder|Input Folder|1"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "file")
        self.assertEqual(param.name(), "in_folder")
        self.assertEqual(param.description(), "Input Folder")
        self.assertEqual(param.behavior(), QgsProcessingParameterFile.Behavior.Folder)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFile|in_folder|Input Folder|QgsProcessingParameterFile.Folder"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "file")
        self.assertEqual(param.name(), "in_folder")
        self.assertEqual(param.description(), "Input Folder")
        self.assertEqual(param.behavior(), QgsProcessingParameterFile.Behavior.Folder)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFile|in_file|Input File|0|gpkg"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "file")
        self.assertEqual(param.name(), "in_file")
        self.assertEqual(param.description(), "Input File")
        self.assertEqual(param.behavior(), QgsProcessingParameterFile.Behavior.File)
        self.assertEqual(param.extension(), "gpkg")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFile|in_file|Input File|0|png|None|False|PNG Files (*.png);; JPG Files (*.jpg *.jpeg)"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "file")
        self.assertEqual(param.name(), "in_file")
        self.assertEqual(param.description(), "Input File")
        self.assertEqual(param.behavior(), QgsProcessingParameterFile.Behavior.File)
        self.assertEqual(param.extension(), "")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertEqual(
            param.fileFilter(), "PNG Files (*.png);; JPG Files (*.jpg *.jpeg)"
        )

    def testParameterVectorDestDesc(self):
        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination")
        self.assertEqual(
            param.dataType(), QgsProcessing.SourceType.TypeVectorAnyGeometry
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination Point|0"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination Point")
        self.assertEqual(param.dataType(), QgsProcessing.SourceType.TypeVectorPoint)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination Point|QgsProcessing.TypeVectorPoint"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination Point")
        self.assertEqual(param.dataType(), QgsProcessing.SourceType.TypeVectorPoint)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination Line|1"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination Line")
        self.assertEqual(param.dataType(), QgsProcessing.SourceType.TypeVectorLine)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination Line|QgsProcessing.TypeVectorLine"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination Line")
        self.assertEqual(param.dataType(), QgsProcessing.SourceType.TypeVectorLine)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination Polygon|2"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination Polygon")
        self.assertEqual(param.dataType(), QgsProcessing.SourceType.TypeVectorPolygon)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination Polygon|QgsProcessing.TypeVectorPolygon"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination Polygon")
        self.assertEqual(param.dataType(), QgsProcessing.SourceType.TypeVectorPolygon)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination Table|5"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination Table")
        self.assertEqual(param.dataType(), QgsProcessing.SourceType.TypeVector)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination Table|QgsProcessing.TypeVector"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination Table")
        self.assertEqual(param.dataType(), QgsProcessing.SourceType.TypeVector)
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterVectorDestination|param_vector_dest|Vector Destination|-1|None|True|False"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vectorDestination")
        self.assertEqual(param.name(), "param_vector_dest")
        self.assertEqual(param.description(), "Vector Destination")
        self.assertEqual(
            param.dataType(), QgsProcessing.SourceType.TypeVectorAnyGeometry
        )
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertFalse(param.createByDefault())

    def testParameterRasterDestDesc(self):
        desc = "QgsProcessingParameterRasterDestination|param_raster_dest|Raster Destination"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "rasterDestination")
        self.assertEqual(param.name(), "param_raster_dest")
        self.assertEqual(param.description(), "Raster Destination")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterRasterDestination|param_raster_dest|Raster Destination|None|True|False"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "rasterDestination")
        self.assertEqual(param.name(), "param_raster_dest")
        self.assertEqual(param.description(), "Raster Destination")
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertFalse(param.createByDefault())

    def testParameterFolderDestDesc(self):
        desc = "QgsProcessingParameterFolderDestination|param_folder_dest|Folder Destination"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "folderDestination")
        self.assertEqual(param.name(), "param_folder_dest")
        self.assertEqual(param.description(), "Folder Destination")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterFolderDestination|param_folder_dest|Folder Destination|None|True|False"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "folderDestination")
        self.assertEqual(param.name(), "param_folder_dest")
        self.assertEqual(param.description(), "Folder Destination")
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertFalse(param.createByDefault())

    def testParameterFileDestDesc(self):
        desc = "QgsProcessingParameterFileDestination|param_file_dest|File Destination"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "fileDestination")
        self.assertEqual(param.name(), "param_file_dest")
        self.assertEqual(param.description(), "File Destination")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterFileDestination|param_html_dest|HTML File Destination|HTML Files (*.html)"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "fileDestination")
        self.assertEqual(param.name(), "param_html_dest")
        self.assertEqual(param.description(), "HTML File Destination")
        self.assertEqual(param.fileFilter(), "HTML Files (*.html)")
        self.assertEqual(param.defaultFileExtension(), "html")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterFileDestination|param_img_dest|Img File Destination|PNG Files (*.png);; JPG Files (*.jpg *.jpeg)"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "fileDestination")
        self.assertEqual(param.name(), "param_img_dest")
        self.assertEqual(param.description(), "Img File Destination")
        self.assertEqual(
            param.fileFilter(), "PNG Files (*.png);; JPG Files (*.jpg *.jpeg)"
        )
        self.assertEqual(param.defaultFileExtension(), "png")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertTrue(param.createByDefault())

        desc = "QgsProcessingParameterFileDestination|param_csv_dest|CSV File Destination|CSV Files (*.csv)|None|True|False"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "fileDestination")
        self.assertEqual(param.name(), "param_csv_dest")
        self.assertEqual(param.description(), "CSV File Destination")
        self.assertEqual(param.fileFilter(), "CSV Files (*.csv)")
        self.assertEqual(param.defaultFileExtension(), "csv")
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )
        self.assertFalse(param.createByDefault())

    def testParameterFeatureSourceDesc(self):
        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(param.dataTypes(), [])
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|0"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorPoint]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|QgsProcessing.TypeVectorPoint"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorPoint]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|1"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorLine]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|QgsProcessing.TypeVectorLine"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorLine]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|2"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorPolygon]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|QgsProcessing.TypeVectorPolygon"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorPolygon]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|5"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(param.dataTypes(), [QgsProcessing.SourceType.TypeVector])
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|QgsProcessing.TypeVector"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(param.dataTypes(), [QgsProcessing.SourceType.TypeVector])
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|1;2"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(),
            [
                QgsProcessing.SourceType.TypeVectorLine,
                QgsProcessing.SourceType.TypeVectorPolygon,
            ],
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|QgsProcessing.TypeVectorLine;QgsProcessing.TypeVectorPolygon"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(),
            [
                QgsProcessing.SourceType.TypeVectorLine,
                QgsProcessing.SourceType.TypeVectorPolygon,
            ],
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterFeatureSource|in_vector|Input Vector|-1|None|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "source")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorAnyGeometry]
        )
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

    def testParameterVectorLayerDesc(self):
        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(param.dataTypes(), [])
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|0"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorPoint]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|QgsProcessing.TypeVectorPoint"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorPoint]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|1"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorLine]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|QgsProcessing.TypeVectorLine"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorLine]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|2"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorPolygon]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|QgsProcessing.TypeVectorPolygon"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorPolygon]
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|5"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(param.dataTypes(), [QgsProcessing.SourceType.TypeVector])
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|QgsProcessing.TypeVector"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(param.dataTypes(), [QgsProcessing.SourceType.TypeVector])
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|1;2"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(),
            [
                QgsProcessing.SourceType.TypeVectorLine,
                QgsProcessing.SourceType.TypeVectorPolygon,
            ],
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|QgsProcessing.TypeVectorLine;QgsProcessing.TypeVectorPolygon"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(),
            [
                QgsProcessing.SourceType.TypeVectorLine,
                QgsProcessing.SourceType.TypeVectorPolygon,
            ],
        )
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterVectorLayer|in_vector|Input Vector|-1|None|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "vector")
        self.assertEqual(param.name(), "in_vector")
        self.assertEqual(param.description(), "Input Vector")
        self.assertListEqual(
            param.dataTypes(), [QgsProcessing.SourceType.TypeVectorAnyGeometry]
        )
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

    def testParameterRasterLayerDesc(self):
        desc = "QgsProcessingParameterRasterLayer|in_raster|Input Raster"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "raster")
        self.assertEqual(param.name(), "in_raster")
        self.assertEqual(param.description(), "Input Raster")
        self.assertIsNone(param.defaultValue())
        self.assertFalse(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )

        desc = "QgsProcessingParameterRasterLayer|in_raster|Input Raster|None|True"
        param = getParameterFromString(desc)
        self.assertIsNotNone(param)
        self.assertEqual(param.type(), "raster")
        self.assertEqual(param.name(), "in_raster")
        self.assertEqual(param.description(), "Input Raster")
        self.assertIsNone(param.defaultValue())
        self.assertTrue(
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
        )


if __name__ == "__main__":
    unittest.main()



"""
***************************************************************************
    RasterCalculatorWidgets.py
    ---------------------
    Date                 : November 2016
    Copyright            : (C) 2016 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "November 2016"
__copyright__ = "(C) 2016, Victor Olaya"

import json
import os
import re
from functools import partial

from qgis.analysis import QgsRasterCalcNode, QgsRasterCalculatorEntry
from qgis.core import (
    QgsProcessingOutputRasterLayer,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterRasterLayer,
    QgsProcessingUtils,
    QgsProject,
)
from qgis.PyQt import uic
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtGui import QTextCursor
from qgis.PyQt.QtWidgets import (
    QComboBox,
    QLabel,
    QLineEdit,
    QListWidgetItem,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
)
from qgis.utils import iface

from processing.core.exceptions import InvalidParameterValue
from processing.gui.BatchInputSelectionPanel import BatchInputSelectionPanel
from processing.gui.wrappers import (
    DIALOG_BATCH,
    DIALOG_STANDARD,
    WidgetWrapper,
)
from processing.tools import dataobjects
from processing.tools.system import userFolder

pluginPath = os.path.dirname(__file__)
WIDGET_ADD_NEW, BASE_ADD_NEW = uic.loadUiType(
    os.path.join(pluginPath, "AddNewExpressionDialog.ui")
)


class AddNewExpressionDialog(BASE_ADD_NEW, WIDGET_ADD_NEW):
    def __init__(self, expression):
        super().__init__()
        self.setupUi(self)

        self.name = None
        self.expression = None
        self.txtExpression.setPlainText(expression)
        self.buttonBox.rejected.connect(self.cancelPressed)
        self.buttonBox.accepted.connect(self.okPressed)

    def cancelPressed(self):
        self.close()

    def okPressed(self):
        self.name = self.txtName.text()
        self.expression = self.txtExpression.toPlainText()
        self.close()


WIDGET_DLG, BASE_DLG = uic.loadUiType(
    os.path.join(pluginPath, "PredefinedExpressionDialog.ui")
)


class PredefinedExpressionDialog(BASE_DLG, WIDGET_DLG):
    def __init__(self, expression, options):
        super().__init__()
        self.setupUi(self)

        self.filledExpression = None
        self.options = options
        self.expression = expression
        self.variables = set(re.findall(r"\[.*?\]", expression))
        self.comboBoxes = {}
        for variable in self.variables:
            label = QLabel(variable[1:-1])
            combo = QComboBox()
            for opt in self.options.keys():
                combo.addItem(opt)
            self.comboBoxes[variable] = combo
            self.groupBox.layout().addWidget(label)
            self.groupBox.layout().addWidget(combo)

        verticalSpacer = QSpacerItem(
            20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
        )
        self.groupBox.layout().addItem(verticalSpacer)

        self.buttonBox.rejected.connect(self.cancelPressed)
        self.buttonBox.accepted.connect(self.okPressed)

    def cancelPressed(self):
        self.close()

    def okPressed(self):
        self.filledExpression = self.expression
        for name, combo in self.comboBoxes.items():
            self.filledExpression = self.filledExpression.replace(
                name, self.options[combo.currentText()]
            )
        self.close()


WIDGET, BASE = uic.loadUiType(os.path.join(pluginPath, "RasterCalculatorWidget.ui"))


class ExpressionWidget(BASE, WIDGET):
    _expressions = {"NDVI": "([NIR] - [Red]) / ([NIR] + [Red])"}

    def __init__(self, options):
        super().__init__(None)
        self.setupUi(self)

        self.setList(options)

        def doubleClicked(item):
            self.text.insertPlainText(f'"{self.options[item.text()]}"')

        def addButtonText(text):
            if any(c for c in text if c.islower()):
                self.text.insertPlainText(f" {text}()")
                self.text.moveCursor(
                    QTextCursor.MoveOperation.PreviousCharacter,
                    QTextCursor.MoveMode.MoveAnchor,
                )
            else:
                self.text.insertPlainText(f" {text} ")

        buttons = [
            b for b in self.buttonsGroupBox.children() if isinstance(b, QPushButton)
        ]
        for button in buttons:
            button.clicked.connect(partial(addButtonText, button.text()))
        self.listWidget.itemDoubleClicked.connect(doubleClicked)

        self.expressions = {}
        if os.path.exists(self.expsFile()):
            with open(self.expsFile()) as f:
                self.expressions.update(json.load(f))
        self.expressions.update(self._expressions)

        self.fillPredefined()
        self.buttonAddPredefined.clicked.connect(self.addPredefined)

        self.buttonSavePredefined.clicked.connect(self.savePredefined)
        self.text.textChanged.connect(self.expressionValid)

    def expressionValid(self):
        errorString = ""
        testNode = QgsRasterCalcNode.parseRasterCalcString(
            self.text.toPlainText(), errorString
        )

        if not self.text.toPlainText():
            self.expressionErrorLabel.setText(self.tr("Expression is empty"))
            self.expressionErrorLabel.setStyleSheet("QLabel { color: black; }")
            return False

        if testNode:
            self.expressionErrorLabel.setText(self.tr("Expression is valid"))
            self.expressionErrorLabel.setStyleSheet(
                "QLabel { color: green; font-weight: bold; }"
            )
            return True

        self.expressionErrorLabel.setText(
            self.tr("Expression is not valid ") + errorString
        )
        self.expressionErrorLabel.setStyleSheet(
            "QLabel { color : red; font-weight: bold; }"
        )
        return False

    def expsFile(self):
        return os.path.join(userFolder(), "rastercalcexpressions.json")

    def addPredefined(self):
        expression = self.expressions[self.comboPredefined.currentText()]
        dlg = PredefinedExpressionDialog(expression, self.options)
        dlg.exec()
        if dlg.filledExpression:
            self.text.setPlainText(dlg.filledExpression)

    def savePredefined(self):
        exp = self.text.toPlainText()
        used = [v for v in self.options.values() if v in exp]

        for i, v in enumerate(used):
            exp = exp.replace(v, f"[{chr(97 + i)}]")

        dlg = AddNewExpressionDialog(exp)
        dlg.exec()
        if dlg.name:
            self.expressions[dlg.name] = dlg.expression

        with open(self.expsFile(), "w") as f:
            f.write(json.dumps(self.expressions))

    def fillPredefined(self):
        self.comboPredefined.clear()
        for expression in self.expressions:
            self.comboPredefined.addItem(expression)

    def setList(self, options):
        self.options = options
        self.listWidget.clear()
        entries = QgsRasterCalculatorEntry.rasterEntries()

        def _find_source(name):
            for entry in entries:
                if entry.ref == name:
                    return entry.raster.source()
            return ""

        for name in options.keys():
            item = QListWidgetItem(name, self.listWidget)
            tooltip = _find_source(name)
            if tooltip:
                item.setData(Qt.ItemDataRole.ToolTipRole, tooltip)
            self.listWidget.addItem(item)

    def setValue(self, value):
        self.text.setPlainText(value)

    def value(self):
        return self.text.toPlainText()


class ExpressionWidgetWrapper(WidgetWrapper):
    def _panel(self, options):
        return ExpressionWidget(options)

    def _get_options(self):
        entries = QgsRasterCalculatorEntry.rasterEntries()
        options = {}
        for entry in entries:
            options[entry.ref] = entry.ref
        return options

    def createWidget(self):
        if self.dialogType == DIALOG_STANDARD:
            if (
                iface is not None
                and iface.layerTreeView() is not None
                and iface.layerTreeView().layerTreeModel() is not None
            ):
                iface.layerTreeView().layerTreeModel().dataChanged.connect(self.refresh)
            return self._panel(self._get_options())
        elif self.dialogType == DIALOG_BATCH:
            return QLineEdit()
        else:
            layers = self.dialog.getAvailableValuesOfType(
                [QgsProcessingParameterRasterLayer], [QgsProcessingOutputRasterLayer]
            )
            options = {
                self.dialog.resolveValueDescription(
                    lyr
                ): f"{self.dialog.resolveValueDescription(lyr)}@1"
                for lyr in layers
            }
            self.widget = self._panel(options)
            return self.widget

    def refresh(self, *args):
        self.widget.setList(self._get_options())

    def setValue(self, value):
        if self.dialogType == DIALOG_STANDARD:
            pass  # TODO
        elif self.dialogType == DIALOG_BATCH:
            return self.widget.setText(value)
        else:
            self.widget.setValue(value)

    def value(self):
        if self.dialogType == DIALOG_STANDARD:
            return self.widget.value()
        elif self.dialogType == DIALOG_BATCH:
            return self.widget.text()
        else:
            return self.widget.value()


class LayersListWidgetWrapper(WidgetWrapper):
    def createWidget(self):
        if self.dialogType == DIALOG_BATCH:
            widget = BatchInputSelectionPanel(
                self.parameterDefinition(), self.row, self.col, self.dialog
            )
            widget.valueChanged.connect(lambda: self.widgetValueHasChanged.emit(self))
            return widget
        else:
            return None

    def setValue(self, value):
        if self.dialogType == DIALOG_BATCH:
            return self.widget.setText(value)

    def value(self):
        if self.dialogType == DIALOG_STANDARD:
            if self.param.datatype == dataobjects.TYPE_FILE:
                return self.param.setValue(self.widget.selectedoptions)
            else:
                if self.param.datatype == dataobjects.TYPE_RASTER:
                    options = QgsProcessingUtils.compatibleRasterLayers(
                        QgsProject.instance(), False
                    )
                elif self.param.datatype == dataobjects.TYPE_VECTOR_ANY:
                    options = QgsProcessingUtils.compatibleVectorLayers(
                        QgsProject.instance(), [], False
                    )
                else:
                    options = QgsProcessingUtils.compatibleVectorLayers(
                        QgsProject.instance(), [self.param.datatype], False
                    )
                return [options[i] for i in self.widget.selectedoptions]
        elif self.dialogType == DIALOG_BATCH:
            return self.widget.getText()
        else:
            options = self._getOptions()
            values = [options[i] for i in self.widget.selectedoptions]
            if (
                len(values) == 0
                and not self.parameterDefinition().flags()
                & QgsProcessingParameterDefinition.Flag.FlagOptional
            ):
                raise InvalidParameterValue(self.parameterDefinition(), self.widget)
            return values



"""
/***************************************************************************
                           qgsplugininstallerrepositorydialog.py
                           Plugin Installer module
                             -------------------
    Date                 : June 2013
    Copyright            : (C) 2013 by Borys Jurgiel
    Email                : info at borysjurgiel dot pl

    This module is based on former plugin_installer plugin:
      Copyright (C) 2007-2008 Matthew Perry
      Copyright (C) 2008-2013 Borys Jurgiel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from pathlib import Path

from qgis.PyQt import uic
from qgis.PyQt.QtWidgets import QDialog, QDialogButtonBox

Ui_QgsPluginInstallerRepositoryDetailsDialogBase, _ = uic.loadUiType(
    Path(__file__).parent / "qgsplugininstallerrepositorybase.ui"
)


class QgsPluginInstallerRepositoryDialog(
    QDialog, Ui_QgsPluginInstallerRepositoryDetailsDialogBase
):
    # ----------------------------------------- #

    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setupUi(self)
        self.editURL.setText("http://")
        self.editName.textChanged.connect(self.textChanged)
        self.editURL.textChanged.connect(self.textChanged)
        self.textChanged(None)

    # ----------------------------------------- #
    def textChanged(self, string):
        enable = len(self.editName.text()) > 0 and len(self.editURL.text()) > 0
        self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(enable)

    def configId(self):
        return self.editAuthCfgWgt.configId()



"""
***************************************************************************
    v_net_distance.py
    ---------------------
    Date                 : December 2015
    Copyright            : (C) 2015 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "December 2015"
__copyright__ = "(C) 2015, Médéric Ribreux"

import os

from processing.tools.system import getTempFilename
from qgis.core import QgsProcessingParameterString

from .v_net import variableOutput


def processCommand(alg, parameters, context, feedback):
    """Handle data preparation for v.net.distance:
    * Integrate point layers into network vector map.
    * Make v.net.distance use those layers.
    * Delete the threshold parameter.
    * If where statement, connect to the db
    """
    # Grab the point layer and delete this parameter
    lineLayer = alg.exportedLayers["input"]
    fromLayer = alg.exportedLayers["flayer"]
    toLayer = alg.exportedLayers["tlayer"]
    intLayer = "bufnet" + os.path.basename(getTempFilename(context=context))
    netLayer = "net" + os.path.basename(getTempFilename(context=context))
    threshold = alg.parameterAsDouble(parameters, "threshold", context)

    # Create the v.net connect command for from_layer integration
    command = f"v.net -s input={lineLayer} points={fromLayer} output={intLayer} operation=connect threshold={threshold} arc_layer=1 node_layer=2"
    alg.commands.append(command)

    # Do it again with to_layer
    command = f"v.net -s input={intLayer} points={toLayer} output={netLayer} operation=connect threshold={threshold} arc_layer=1 node_layer=3"
    alg.commands.append(command)

    # Connect the point layer database to the layer 2 of the network
    command = f"v.db.connect -o map={netLayer} table={fromLayer} layer=2"
    alg.commands.append(command)

    command = f"v.db.connect -o map={netLayer} table={toLayer} layer=3"
    alg.commands.append(command)

    # remove undesired parameters
    alg.removeParameter("flayer")
    alg.removeParameter("tlayer")
    alg.removeParameter("threshold")
    alg.exportedLayers["input"] = netLayer

    # Add the two new parameters
    fLayer = QgsProcessingParameterString("from_layer", None, 2, False, False)
    alg.addParameter(fLayer)
    tLayer = QgsProcessingParameterString("to_layer", None, 3, False, False)
    alg.addParameter(tLayer)
    alg.processCommand(parameters, context, feedback)


def processOutputs(alg, parameters, context, feedback):
    outputParameter = {"output": ["output", "line", 1, True]}
    variableOutput(alg, outputParameter, parameters, context)
